package table

import org.assertj.core.api.Assertions.assertThat
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.TestInstance.Lifecycle.PER_CLASS
import sexpression.Nil.NIL
import sexpression.Symbol.T
import table.SymbolTable.NullSymbolTable

@TestInstance(PER_CLASS)
class ExecutionContextTest {

    private val executionContext: ExecutionContext = ExecutionContext

    @BeforeEach
    fun setUp() {
        executionContext.clearContext()
    }

    @AfterEach
    fun tearDown() {
        executionContext.clearContext()
    }

    @Test
    fun `assign a new scope`() {
        val scope = SymbolTable()
        executionContext.scope = scope

        assertThat(executionContext.scope).isEqualTo(scope)
    }

    @Test
    fun `clear the context`() {
        val scope = SymbolTable()
        executionContext.scope = scope

        assertThat(executionContext.scope).isEqualTo(scope)
        executionContext.clearContext()
        assertThat(executionContext.scope).isNotEqualTo(scope)
        assertThat(executionContext.scope.parent).isEqualTo(NullSymbolTable)
    }

    @Test
    fun `lookup a variable`() {
        executionContext.scope["test"] = T

        assertThat(executionContext.lookupSymbolValue("test")).isEqualTo(T)
    }

    @Test
    fun `lookup a local variable`() {
        val scope = SymbolTable(executionContext.scope)

        scope["local"] = T
        executionContext.scope = scope

        assertThat(executionContext.lookupSymbolValue("local")).isEqualTo(T)
    }

    @Test
    fun `lookup a global variable`() {
        val global = executionContext.scope
        val scope1 = SymbolTable(global)
        val scope2 = SymbolTable(scope1)
        val scope3 = SymbolTable(scope2)
        executionContext.scope["global"] = T
        executionContext.scope = scope3

        assertThat(executionContext.lookupSymbolValue("global")).isEqualTo(T)
    }

    @Test
    fun `lookup a shadowed variable`() {
        val scope = SymbolTable(executionContext.scope)

        scope["shadowed"] = NIL
        executionContext.scope["shadowed"] = T
        executionContext.scope = scope

        assertThat(executionContext.lookupSymbolValue("shadowed")).isEqualTo(NIL)
    }

    @Test
    fun `restore the global context`() {
        val global = executionContext.scope
        val scope1 = SymbolTable(global)
        val scope2 = SymbolTable(scope1)
        val scope3 = SymbolTable(scope2)
        executionContext.scope = scope3

        assertThat(executionContext.scope.isGlobal).isFalse()
        executionContext.restoreGlobalScope()
        assertThat(executionContext.scope.isGlobal).isTrue()
        assertThat(executionContext.scope).isEqualTo(global)
    }
}
